/**
* Function: dbo.napgen__make_uuid
*
* Description: Create a version 5 (i.e. SHA-1-based) UUID of the
*              specified namespace-qualified name.
*
*              The RFC is at:
*
*                  http://tools.ietf.org/html/rfc4122
*
*              The code here is based on the one at:
*
*                  http://www.php.net/manual/de/function.uniqid.php#94959
*
* @param varchar(max) s_ns_uuid The UUID of a namespace. Blank means
*                               there is no namespace.
*
* @param varchar(max) s_lcl_name The local name (local to the
*                                specified namespace UUID) for which
*                                an UUID is to be created.
*
* @return varchar(36) A version 5 UUID.
*
* @example To generate the UUID of the namespace 'nsTest', in SQL
           Server Management Studio's SQL query window do:
*
*              select dbo.napgen__make_uuid('', 'nsTest');
*              go
*
*          The UUID should be:
*
*              c4e1a961-f4f3-52db-977d-721685de631c
*
* @example To generate the UUID of the name 'test' in the namespace
*          'nsTest', in SQL Server Management Studio's SQL query
*          window do:
*
*              select dbo.napgen__make_uuid(dbo.napgen__make_uuid('', 'nsTest'), 'test');
*              go
*          or
*              select dbo.napgen__make_uuid('c4e1a961-f4f3-52db-977d-721685de631c', 'test');
*              go
*
*          In both cases, the UUID should be:
*
*              fe55d30d-ed14-5d37-b4c2-c566e739b919
*
* @useCase Adding UUID into an existing database.
*
*    Notes about the example setting:
*
*        - In the database, the table holding the main record is
*          called "main_tbl".
*
*        - To add meaning to the namespace, we name it using the main
*          table's name, i.e. "main_tbl". To avoid collision with
*          another namespace which happens to be also called
*          "main_tbl", we tag on a bracketed time stamp (in this case
*          "2012/12/06:10:20:30"). One can use a random number
*          instead.
*
*        - To generate an UUID specific to a record, we use the
*          record's primary key (from the rec_id column) as a name.
*
*        - The record UUID is stored in the rec_uuid column.
*
*    To add a column holding a record's UUID:
*
*        alter table main_tbl
*        add rec_uuid varchar(36) unique;
*
*    To fill the rec_uuid column:
*
*        update main_tbl
*        set rec_uuid =
*            dbo.napgen__make_uuid(dbo.napgen__make_uuid('',
*                                                        'main_tbl(2012/12/06:10:20:30)'),
*                                  cast(rec_id as varchar(10)));
*
*    A trigger that generates a record UUID when a record is inserted
*    into a database table:
*
*        create trigger trig_insert_main_tbl
*            on main_tbl
*            after insert as
*        begin
*            update main_tbl
*            set rec_uuid =
*                dbo.napgen__make_uuid(dbo.napgen__make_uuid('',
*                                                            'main_tbl(2012/12/06:10:20:30)'),
*                                      cast(rec_id as varchar(10)))
*            where rec_uuid is null;
*        end;
*
* @installation To install this function, do the following:
*
*     1. Use SQL Server Management Studio to connect to a database
*        server.
*
*     2. Start an SQL query window by clicking on the "New Query"
*        button.
*
*     3. Allow stored functions to use OLE (Object Linking and
*        Embedding) by executing the following:
*
*            exec sp_configure 'show advanced options', 1;
*            go
*            reconfigure;
*            go
*            exec sp_configure 'Ole Automation Procedures';
*            go
*            exec sp_configure 'Ole Automation Procedures', 1;
*            go
*            reconfigure;
*            go
*
*     4. Load and execute this file.
*
*     5. Execute the following:
*
*            select dbo.napgen__make_uuid(dbo.napgen__make_uuid('', 'nsTest'),'test');
*            go
*
*     6. You should have the result "fe55d30d-ed14-5d37-b4c2-c566e739b919".
*
* @author Boniface Lau (boniface.lau@gmail.com)
*/
if object_id('dbo.napgen__make_uuid', 'FN') is not null
    drop function dbo.napgen__make_uuid;
go

create function dbo.napgen__make_uuid(@s_ns_uuid varchar(max),
                                      @s_lcl_name varchar(max))
                returns varchar(36) as
begin
    declare
        @s_ns_uuid_hex as varchar(-- A hex-only UUID always has 32 hex
                                  -- characters.
                                  32),
        @s_lcl_name_hex varchar(max),
        @s_hash_hex nvarchar(-- SHA-1 is a 160-bit hash. Thus,
                             -- requires 40 hex characters to
                             -- represent 20 8-bit bytes.
                             40),
        @oRe int,
        @iRet int,
        @iReMatch bit,
        @i_time_hi_and_version varbinary(-- A 16-bit field.
                                         2),
        @i_clock_seq_hi_and_reserved varbinary(-- An 8-bit field.
                                               1);

    set @s_ns_uuid_hex = '';
    if @s_ns_uuid != ''
        -- User did specify something as a namespace UUID.

        begin
        -- T-SQL does not have regular expression and its string
        -- manipulation capability is very primitive. We therefore
        -- have to resort to OLE to access VBScript's regular
        -- expression engine (it implements regular expression
        -- according to the ECMA standard). The engine is distributed
        -- with Internet Explorer and therefore should be available on
        -- all Windows platforms.
        exec @iRet = sp_OACreate 'VBScript.RegExp', @oRe out;
        if @iRet = 0
            -- The regular expression object has been created. Set it
            -- up to check the format of the specified namespace UUID.
            begin
            exec @iRet = sp_OASetProperty @oRe,
                                          'Pattern',
                                          '^\{?[0-9a-f]{8}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{12}\}?$';
            if @iRet = 0
                begin
                exec @iRet = sp_OASetProperty @oRe,
                                              'IgnoreCase',
                                              1;
                if @iRet = 0
                    begin
                    exec @iRet = sp_OAMethod @oRe,
                                             'Test',
                                             @iReMatch OUT,
                                             @s_ns_uuid;
                    if @iRet = 0
                        begin
                        if @iReMatch = 1
                            -- The specified namespace UUID is in a
                            -- valid format.

                            begin
                            -- Keep only the UUID's hex digits. We need them to construct a binary
                            -- representation of the UUID.
                            set @s_ns_uuid_hex = -- Drop all the noise (non-hex) characters from the
                                                 -- UUID.
                                                 replace(replace(replace(@s_ns_uuid,
                                                                         '{', ''),
                                                                 '}', ''),
                                                         '-', '');
                            end
                        else
                            begin
                            -- T-SQL does not allow a user defined function to raise error through
                            -- raiserror(). For lack of a better option on forcing an abort, we
                            -- trigger an error through invalid type casting. The message that we
                            -- want to display will be visible to the user. This can be confusing.
                            -- But we are constrained by T-SQL.
                            return cast('The namespace UUID "' +
                                                         @s_ns_uuid +
                                                           '" does not have the right format.'
                                        as int);
                            end
                        end
                    else
                        begin
                        return cast('Failed to invoke the regular expression method: Test.'
                                    as int);
                        end
                    end
                else
                    begin
                    return cast('Failed to set the regular expression property: IgnoreCase .'
                                as int);
                    end
                end
            else
                begin
                return cast('Failed to set the required regular expression.'
                            as int);
                end
            exec sp_OADestroy @oRe;
            end
        else
            begin
            return cast('Failed to use OLE to access VBScript.RegExp (required for doing regular expression match).'
                        as int);
            end
        end

    set @s_lcl_name_hex = -- Hex representation of the local name's binary.
                          convert(varchar(max),

                                  -- Binary of the local name.
                                  convert(varbinary(max), @s_lcl_name, 0),
                                  2);

    set @s_hash_hex = -- Hex representation of the hash.
                      convert(varchar(40),

                              -- Hash of the binary of the UUID-prefixed local name.
                              hashbytes('sha1',

                                        -- Binary of the UUID-prefixed local name.
                                        convert(varbinary(max),

                                                -- Prefix the local name with the namespace UUID.
                                                @s_ns_uuid_hex + @s_lcl_name_hex,
                                                2)),
                              2);

    -- Set the 16-bit RFC "time_hi_and_version" field using byte 6 and
    -- 7 from the hash. Bit 12 through 15 are for the UUID version
    -- number, in this case version 5.
    set @i_time_hi_and_version = cast(convert(varbinary(2),
                                              substring(@s_hash_hex, 13, 4),
                                              2)
                                      as int)
                                 & 4095   -- Hex '0fff'
                                 | 20480; -- Hex '5000'

    -- Set the 8-bit RFC "clock_seq_hi_and_reserved" field using byte
    -- 8 from the hash. The two most significant bits are set to 0 and
    -- 1 indicating the value "DCE1.1".
    set @i_clock_seq_hi_and_reserved = cast(convert(varbinary(1),
                                                    substring(@s_hash_hex, 17, 2),
                                                    2)
                                            as int)
                                       & 63   -- Hex '3f'
                                       | 128; -- Hex '80'

    -- Convert the hash to an UUID according to the RFC.
    return -- Set the 32-bit RFC "time_low" field using the first
           -- eight hex digits (byte 0 to 3) from the hash.
           substring(@s_hash_hex, 1, 8) + '-' +

           -- Set the 16-bit RFC "time_mid" field using the next four
           -- hex digits (byte 4 and 5) from the hash.
           substring(@s_hash_hex, 9, 4) + '-' +

           -- The 2-byte "time_hi_and_version" field.
           right(-- Using XQuery's hexBinary() to do hex conversion is
                 -- much cleaner than using T-SQL's string
                 -- manipulation. XQuery was available since SQL Server
                 -- 2005.
                 cast('' as xml)
                 .value('xs:hexBinary(sql:variable("@i_time_hi_and_version"))',
                        'varchar(max)'),
                 4) + '-' +

           -- The 1-byte "clock_seq_hi_and_reserved" field.
           right(cast('' as xml)
                 .value('xs:hexBinary(sql:variable("@i_clock_seq_hi_and_reserved"))',
                        'varchar(max)'),
                 2) +

           -- Set the 8-bit RFC "clock_seq_low" field using byte 9
           -- from the hash.
           substring(@s_hash_hex, 19, 2) + '-' +

           -- Set the 48-bit RFC "node" field using byte 10 through 15
           -- from the hash.
           substring(@s_hash_hex, 21, 12);

end
go
