create or replace package napgen as
/**
* Package: napgen
*
* Description: Oracle enhancement package from NapGen.
*
* @installation To install this package, do the following:
*
*     1. Enable the pgcrypto module
*          A. Start an Oracle SQL*Plus command line window (e.g.
*             "sqlplus.exe")
*
*          B. Connect as sysdba:
*
*             "conn / as sysdba"
*
*          C. Display the dictionary accessibility setting:
*
*             "show parameter o7_dictionary_accessibility name type value"
*
*          D. If the value is false, do the following:
*              I. Set it to true so that Oracle will allow "sys" to
*                 connect:
*
*                 "alter system set o7_dictionary_accessibility=true scope=spfile;"
*
*              II. Restart Oracle to have the change in effect:
*                  1. "shutdown immediate"
*
*                  2. "startup"
*
*          E. Grant the execute access of the package sys.dbms_crypto to public:
*              I. Connect as sys:
*
*                 "connect sys"
*
*              II. Grant execute access to the public:
*
*                 "grant execute on sys.dbms_crypto to public;"
*
*      2. Add server-side function
*          A. Start an Oracle SQL*Plus command line window (e.g.
*             "sqlplus.exe") and connect to a database.
*
*          B. Execute this file (e.g. "@sql_oracle_make_uuid.sql").
*
*          C. Execute the following query:
*
*             select napgen.make_uuid(napgen.make_uuid('', 'nsTest'),'test') from dual;
*
*          D. You should have the result "fe55d30d-ed14-5d37-b4c2-c566e739b919".
*
* @author Boniface Lau (boniface.lau@gmail.com)
*/

    function make_uuid(s_ns_uuid varchar2,
                       s_lcl_name varchar2) return varchar2;
end napgen;
/

create or replace package body napgen as

    /**
    * Function: make_uuid
    *
    * Description: Create a version 5 (i.e. SHA-1-based) UUID of the
    *              specified namespace-qualified name.
    *
    *              The RFC is at:
    *
    *                  http://tools.ietf.org/html/rfc4122
    *
    *              The code here is based on the one at:
    *
    *                  http://www.php.net/manual/de/function.uniqid.php#94959
    *
    * @param varchar2 s_ns_uuid The UUID of a namespace. Blank means
    *                           there is no namespace.
    *
    * @param varchar2 s_lcl_name The local name (local to the
    *                            specified namepsace UUID) for which
    *                            an UUID is to be created.
    *
    * @return varchar2 A version 5 UUID.
    *
    * @example To generate the UUID of the namespace 'nsTest', in
    *          an Oracle SQL*Plus command line window do:
    *
    *              select napgen.make_uuid('', 'nsTest') from dual;
    *
    *          The UUID should be:
    *
    *              c4e1a961-f4f3-52db-977d-721685de631c
    *
    * @example To generate the UUID of the local name 'test' in the
    *          namespace 'nsTest', in an Oracle SQL*Plus command line
    *          window do:
    *
    *              select napgen.make_uuid(napgen.make_uuid('', 'nsTest'), 'test') from dual;
    *          or
    *              select napgen.make_uuid('c4e1a961-f4f3-52db-977d-721685de631c', 'test') from dual;
    *
    *          In both cases, the UUID should be:
    *
    *              fe55d30d-ed14-5d37-b4c2-c566e739b919
    *
    * @useCase Adding UUID into an existing database.
    *
    *    Notes about the example setting:
    *
    *        - In the database, the table holding the main record is
    *          called "main_tbl".
    *
    *        - To add meaning to the namespace, we name it using the
    *          main table's name, i.e. "main_tbl". To avoid collision
    *          with another namespace which happens to be also called
    *          "main_tbl", we tag on a bracketed time stamp (in this
    *          case "2012/12/06:10:20:30"). One can use a random
    *          number instead.
    *
    *        - To generate an UUID specific to a record, we use the
    *          record's primary key (from the rec_id column) as a
    *          name.
    *
    *        - The record UUID is stored in the rec_uuid column.
    *
    *    To add a column holding a record's UUID:
    *
    *        alter table main_tbl
    *        add (rec_uuid varchar2(36) unique);
    *
    *    To fill the rec_uuid column:
    *
    *        update main_tbl
    *        set rec_uuid =
    *            napgen.make_uuid(napgen.make_uuid('',
    *                                              'main_tbl(2012/12/06:10:20:30)'),
    *                             cast(rec_id as varchar2(10)));
    *
    *    A trigger that generates a record UUID when a record is
    *    inserted into a database table:
    *
    *        create or replace trigger trig_insert_main_tbl
    *            before insert
    *            on main_tbl
    *            for each row
    *        begin
    *          :new.rec_uuid :=
    *              napgen.make_uuid(napgen.make_uuid('',
    *                                                'main_tbl(2012/12/06:10:20:30)'),
    *                               cast(:new.rec_id as varchar2(10)));
    *        end;
    *
    * @author Boniface Lau (boniface.lau@gmail.com)
    */
    function make_uuid(s_ns_uuid varchar2,
                       s_lcl_name varchar2)
             return varchar2 is

        s_ns_uuid_hex varchar2(-- After removing all the noise
                               -- characters from an UUID, we will
                               -- have only 32 hex characters.
                               32) := -- When there is no namespace
                                      -- UUID specified.
                                      '';

        s_hash_hex varchar2(-- SHA-1 is a 160-bit hash. Thus, requires
                            -- 40 hex characters to represent 20 8-bit
                            -- bytes.
                            40) := '';
    begin
        if s_ns_uuid is not null then
            -- User did specify something as a namespace UUID.

            if regexp_count(s_ns_uuid,
                            '^\{?[0-9a-f]{8}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{12}\}?$',
                            1,
                            'i') = 1 then
                -- The specified namespace UUID is in a valid format.

                -- Keep only the hex digits by removing all the noise characters. That will allow
                -- us to construct a binary representation of the UUID.
                s_ns_uuid_hex := regexp_replace(s_ns_uuid, '[\{\}\-]+', '', 1, 0);
            else
                -- The specified namespace UUID is not in the proper
                -- format.
                raise_application_error(-20000,
                                        'The namespace UUID "' ||
                                                         s_ns_uuid ||
                                                           '" does not have the right format.');
            end if;
        end if;

        s_hash_hex :=
            rawtohex(-- Hash of UUID-prefixed local name.
                     sys.dbms_crypto.hash(-- Binary of the UUID-prefixed local name.
                                          hextoraw(-- Prefix the local name with the namespace UUID.
                                                   s_ns_uuid_hex
                                                   ||
                                                   rawtohex(UTL_I18N.STRING_TO_RAW(s_lcl_name,
                                                                                   'AL32UTF8'))),

                                          -- 160-bit SHA1 (Secure Hash Algorithm) hash.
                                          dbms_crypto.hash_sh1));

        -- Convert the hash to an UUID according to the RFC.
        return -- Set the 32-bit RFC "time_low" field using the first
               -- eight hex digits (byte 0 to 3) from the hash.
               substr(s_hash_hex, 1, 8) || '-' ||

               -- Set the 16-bit RFC "time_mid" field using the next four
               -- hex digits (byte 4 and 5) from the hash.
               substr(s_hash_hex, 9, 4) || '-' ||

               -- Set the 16-bit RFC "time_hi_and_version" field using the
               -- next four hex digits (byte 6 and 7) from the hash. Bit
               -- 12 through 15 are for the UUID version number, in this
               -- case version 5.
              utl_raw.bit_or(utl_raw.bit_and(hextoraw(substr(s_hash_hex, 13, 4)),
                                             hextoraw('0fff')),
                             hextoraw('5000')) || '-' ||

               -- Set the 8-bit RFC "clock_seq_hi_and_reserved" field
               -- using the next two hex digits (byte 8) from the hash.
               -- The two most significant bits are set to 0 and 1
               -- indicating the value "DCE1.1".
               utl_raw.bit_or(utl_raw.bit_and(hextoraw(substr(s_hash_hex, 17, 2)),
                                              hextoraw('3f')),
                              hextoraw('80')) ||

               -- Set the 8-bit RFC "clock_seq_low" field using the next
               -- two hex digits (byte 9) from the hash.
               substr(s_hash_hex, 19, 2) || '-' ||

               -- Set the 48-bit RFC "node" field using the next 12 hex
               -- digits (byte 10 through 15) from the hash.
               substr(s_hash_hex, 21, 12);

    end make_uuid;
end napgen;
/

