/**
* Function: napgen__make_uuid
*
* Description: Create a version 5 (i.e. SHA-1-based) UUID of the
*              specified namespace-qualified name.
*
*              The RFC is at:
*
*                  http://tools.ietf.org/html/rfc4122
*
*              The code here is based on the one at:
*
*                  http://www.php.net/manual/de/function.uniqid.php#94959
*
* @param text s_ns_uuid The UUID of a namespace. Blank means there is
*                       no namespace.
*
* @param text s_lcl_name The local name (local to the specified
*                        namespace UUID) for which an UUID is to be
*                        created.
*
* @return text A version 5 UUID.
*
* @example To generate the UUID of the namespace 'nsTest', in
*          PostgreSQL's administartion tool pgAdmin's SQL query window
*          do:
*
*              select napgen__make_uuid('', 'nsTest');
*
*          The UUID should be:
*
*              c4e1a961-f4f3-52db-977d-721685de631c
*
* @example To generate the UUID of the name 'test' in the namespace
*          'nsTest', in PostgreSQL's administartion tool pgAdmin's SQL
*          query window do:
*
*              select napgen__make_uuid(napgen__make_uuid('', 'nsTest'), 'test');
*          or
*              select napgen__make_uuid('c4e1a961-f4f3-52db-977d-721685de631c', 'test');
*
*          In both cases, the UUID should be:
*
*              fe55d30d-ed14-5d37-b4c2-c566e739b919
*
* @useCase Adding UUID into an existing database.
*
*    Notes about the example setting:
*
*        - In the database, the table holding the main record is
*          called "main_tbl".
*
*        - To add meaning to the namespace, we name it using the main
*          table's name, i.e. "main_tbl". To avoid collision with
*          another namespace which happens to be also called
*          "main_tbl", we tag on a bracketed time stamp (in this case
*          "2012/12/06:10:20:30"). One can use a random number
*          instead.
*
*        - To generate an UUID specific to a record, we use the
*          record's primary key (from the rec_id column) as a name.
*
*        - The record UUID is stored in the rec_uuid column.
*
*    To add a column holding a record's UUID:
*
*        alter table main_tbl
*        add column rec_uuid text unique;
*
*    To fill the rec_uuid column:
*
*        update main_tbl
*        set rec_uuid =
*            napgen__make_uuid(napgen__make_uuid('',
*                                                'main_tbl(2012/12/06:10:20:30)'),
*                              rec_id::text);
*
*    A trigger that generates a record UUID when a record is inserted
*    into a database table:
*
*        create or replace function tfn_set_rec_uuid() returns trigger as
*        $$
*        begin
*            new.rec_uuid :=
*                napgen__make_uuid(napgen__make_uuid('',
*                                                    'main_tbl(2012/12/06:10:20:30)'),
*                                  new.rec_id::text);
*            return new;
*        end;
*        $$
*        language plpgsql volatile;
*
*        create trigger trig_insert_main_tbl
*            before insert
*            on main_tbl
*            for each row
*            execute procedure tfn_set_rec_uuid();
*
* @installation To install this function, do the following:
*
*     1. Enable the pgcrypto module. Depending on the PostgreSQL
*        version (as returned by the SQL query "select version()"), do
*        one of the following:
*
*          - PostgreSQL 9.1+ database:
*             A. Connect to the database via PostgreSQL's
*                administartion tool pgAdmin.
*
*             B. In the "Object browser" window, expand to the
*                database in mind.
*
*             C. Then expand on the entry "Extensions".
*
*             D. If the list of extensions does not include pgcrypto,
*                do the following:
*                 I. With the desired database selected, open up an
*                    SQL Editor.
*
*                 II. Execute the following command:
*
*                     create extension pgcrypto;
*
*          - PostgreSQL pre-9.1 database:
*             A. Change the default directory to PostgreSQL's
*                installation binary directory (e.g. "cd
*                \bin\postgreSql-8.3.4\bin")
*
*             B. Load the pgcrypto module to a database through the
*                superuser "postgres". For example:
*
*                psql -U postgres -d my_database < ..\share\contrib\pgcrypto.sql
*
*                Note: The pgcrypto module is usually in PostgreSQL's
*                      contribution folder (share\contrib).
*
*     2. Enable the language plpgsql
*         A. Connect to the database via PostgreSQL's administartion
*            tool pgAdmin.
*
*         B. With the desired database selected, open up an SQL
*            Editor.
*
*         C. Execute the following command:
*
*            select lanname from pg_language;
*
*         D. If the list of languages does not include plpgsql,
*            execute the following command:
*
*            create language plpgsql;
*
*     3. Add server-side function
*         A. Connect to the database via PostgreSQL's administartion
*            tool pgAdmin.
*
*         B. With the desired database selected, open up an SQL
*            Editor.
*
*         C. Load and execute this file.
*
*         D. Execute the following query:
*
*            select napgen__make_uuid(napgen__make_uuid('', 'nsTest'),'test');
*
*         E. You should have the result "fe55d30d-ed14-5d37-b4c2-c566e739b919".
*
* @author Boniface Lau (boniface.lau@gmail.com)
*/
create or replace function napgen__make_uuid(s_ns_uuid text, s_lcl_name text)
                           returns text as
$$
declare
    ba_ns_uuid bytea := ''; -- Binary (as bytes) of the namespace UUID.
    ba_lcl_name bytea := ''; -- Binary (as bytes) of the local name.
    s_char text;
    s_hash text;
begin
    if s_ns_uuid != '' then
        -- User did specify soemething as a namespace UUID.

        if (s_ns_uuid ~*
            E'^\{?[0-9a-f]{8}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{4}\-?[0-9a-f]{12}\}?$') then
            -- The specified namespace UUID is in a valid format.

            -- Convert the hex representation to binary.
            ba_ns_uuid = decode(-- Drop the noise (non-hex) characters. That will allow us to
                                -- construct a binary representation of the UUID.
                                regexp_replace(s_ns_uuid, E'[\{\}\-]+', '', 'g'),

                                'hex');
        else
            -- The specified namespace UUID is not in the proper
            -- format.
            raise exception
                  'The namespace UUID "%" does not have the right format.',
                  s_ns_uuid;
        end if;
    end if;

    -- Convert the local name to binary.
    for i in 1..length(s_lcl_name) loop
        ba_lcl_name = ba_lcl_name || chr(ascii(substr(s_lcl_name, i, 1)))::bytea;
    end loop;

    s_hash = -- Hex representation of the hash.
             encode(-- Hash of the UUID-prefixed local name.
                    digest(-- Prefix the local name with the namespace UUID.
                           ba_ns_uuid || ba_lcl_name,

                           -- 160-bit SHA1 (Secure Hash Algorithm) hash.
                           'sha1'),
                    'hex');

    -- Convert the hash to an UUID according to the RFC.
    return -- Set the 32-bit RFC "time_low" field using the first
           -- eight hex digits (byte 0 to 3) from the hash.
           substr(s_hash, 1, 8) || '-' ||

           -- Set the 16-bit RFC "time_mid" field using the next four
           -- hex digits (byte 4 and 5) from the hash.
           substr(s_hash, 9, 4) || '-' ||

           -- Set the 16-bit RFC "time_hi_and_version" field using the
           -- next four hex digits (byte 6 and 7) from the hash. Bit
           -- 12 through 15 are for the UUID version number, in this
           -- case version 5.
           to_hex((('x' || substr(s_hash, 13, 4))::bit(16) & x'0fff' | x'5000')::int) || '-' ||

           -- Set the 8-bit RFC "clock_seq_hi_and_reserved" field
           -- using the next two hex digits (byte 8) from the hash.
           -- The two most significant bits are set to 0 and 1
           -- indicating the value "DCE1.1".
           to_hex((('x' || substr(s_hash, 17, 2))::bit(8) & x'3f' | x'80')::int) ||

           -- Set the 8-bit RFC "clock_seq_low" field using the next
           -- two hex digits (byte 9) from the hash.
           substr(s_hash, 19, 2) || '-' ||

           -- Set the 48-bit RFC "node" field using the next 12 hex
           -- digits (byte 10 through 15) from the hash.
           substr(s_hash, 21, 12);
end;
$$
language plpgsql;

